@@ -17,6 +17,7 @@ import datetime
 import io
 import locale
 import os
+import re
 import shutil
 import subprocess
 import sys
@@ -89,6 +90,8 @@ class Invoker:
                                          if k in spec.envWhiteList }
         self.__logFileName = None if noLogFiles else spec.logFile
         self.__logFile = DEVNULL
+        self.__makeFds = []
+        self.__makeJobs = None
         self.__trace = trace
         self.__sandboxHelperPath = None
         self.__stdioBuffer = io.BytesIO() if redirect else None
@@ -125,12 +128,13 @@ class Invoker:
             self.__stdoutStream = Unbuffered(sys.stdout.buffer) if showStdOut else DEVNULL
             self.__stderrStream = Unbuffered(sys.stderr.buffer) if showStdErr else DEVNULL
 
-    def __openLog(self):
+    def __openLog(self, header=None):
         # Create log file
         if self.__logFileName:
             self.__logFile = open(self.__logFileName, "ab", buffering=0)
-            self.__logFile.write("### START: {}\n"
-                .format(datetime.datetime.now().ctime())
+            self.__logFile.write("### START: {}{}\n"
+                .format(datetime.datetime.now().ctime(),
+                        (" (" + header + ")") if header else "")
                 .encode(locale.getpreferredencoding()))
         else:
             self.__logFile = DEVNULL
@@ -167,15 +171,14 @@ class Invoker:
 
     async def __runCommand(self, args, cwd, stdout=None, stderr=None,
                            check=False, env=None, universal_newlines=True,
-                           **kwargs):
+                           specEnv=True, **kwargs):
         cmd = " ".join(quote(a) for a in args)
         self.trace(cmd)
-        if env is None:
-            env = self.__env
-        else:
-            _env = self.__env.copy()
-            _env.update(env)
-            env = _env
+
+        _env = self.__env.copy()
+        if specEnv: _env.update(self.__spec.env)
+        if env is not None: _env.update(env)
+        env = _env
 
         if stdout == True:
             # If stdout should be captured we use a dedicated buffer. This
@@ -224,7 +227,7 @@ class Invoker:
                 *args,
                 stdin=self.__stdin, stdout=stdoutRedir, stderr=stderrRedir,
                 env=env, cwd=cwd,
-                **kwargs)
+                **kwargs, pass_fds=self.__makeFds)
         except OSError as e:
             self.fail(str(e), returncode=127)
 
@@ -305,6 +308,16 @@ class Invoker:
             elif clean and mode != InvocationMode.SHELL:
                 emptyDirectory(self.__spec.workspaceWorkspacePath)
 
+            if len(self.__makeFds) == 2:
+                makeFlags = self.__spec.env.get("MAKEFLAGS")
+                if makeFlags is not None:
+                    makeFlags = re.sub(r'-j\s*[0-9]*', '', makeFlags)
+                    makeFlags = re.sub(r'--jobserver-auth=[0-9]*,[0-9]*', '', makeFlags)
+                else:
+                    makeFlags = ""
+                self.__spec.env["MAKEFLAGS"] = (makeFlags + " -j" + str(self.__makeJobs)
+                    + " --jobserver-auth=" + ",".join([str(fd) for fd in self.__makeFds]))
+
             # setup script and arguments
             if mode == InvocationMode.SHELL:
                 realScriptFile, execScriptFile, callArgs = self.__spec.language.setupShell(
@@ -325,8 +338,9 @@ class Invoker:
                 if not self.__spec.sandboxNetAccess: cmdArgs.append('-n')
 
                 # Create empty env file. Otheriwse bind mount fails.
-                with open(self.__spec.envFile, "wb"): pass
-                cmdArgs.extend(["-M", os.path.abspath(self.__spec.envFile), "-w", "/bob/env"])
+                if self.__spec.envFile:
+                    with open(self.__spec.envFile, "wb"): pass
+                    cmdArgs.extend(["-M", os.path.abspath(self.__spec.envFile), "-w", "/bob/env"])
 
                 # Mount workspace writable and all dependencies read-only
                 cmdArgs.extend(["-M", os.path.abspath(self.__spec.workspaceWorkspacePath),
@@ -343,7 +357,7 @@ class Invoker:
             cmdArgs.extend(callArgs)
 
             if mode == InvocationMode.SHELL:
-                ret = await self.callCommand(cmdArgs)
+                ret = await self.callCommand(cmdArgs, specEnv=False)
             elif mode == InvocationMode.CALL:
                 for scm in self.__spec.preRunCmds:
                     scm = getScm(scm)
@@ -356,7 +370,7 @@ class Invoker:
                     except Exception:
                         self.error(scm.getSource(), "failed")
                         raise
-                await self.checkCommand(cmdArgs)
+                await self.checkCommand(cmdArgs, specEnv=False)
                 for a in self.__spec.postRunCmds:
                     a = CheckoutAssert(a)
                     try:
@@ -388,6 +402,50 @@ class Invoker:
 
         return ret
 
+    async def executeLocalSCMs(self):
+        # make permissions predictable
+        os.umask(0o022)
+
+        ret = 1
+        try:
+            self.__openLog("only local SCMs")
+
+            # execute only local SCMs
+            for scm in self.__spec.preRunCmds:
+                scm = getScm(scm)
+                if not scm.isLocal(): continue
+                try:
+                    await scm.invoke(self)
+                except CmdFailedError as e:
+                    self.error(scm.getSource(), "failed")
+                    self.error(e.what)
+                    raise
+                except Exception:
+                    self.error(scm.getSource(), "failed")
+                    raise
+
+            # still run all checkout asserts afterwards
+            for a in self.__spec.postRunCmds:
+                a = CheckoutAssert(a)
+                try:
+                    await a.invoke(self)
+                except Exception:
+                    self.error(a.getSource(), "failed")
+                    raise
+
+            # everything went well
+            ret = 0
+
+        except OSError as e:
+            self.error("Something went wrong:", str(e))
+            ret = 1
+        except InvocationError as e:
+            ret = e.returncode
+        finally:
+            self.__closeLog(ret)
+
+        return ret
+
     async def executeFingerprint(self, keepSandbox=False):
         # make permissions predictable
         os.umask(0o022)
@@ -397,7 +455,7 @@ class Invoker:
         ret = -1
         stdout = stderr = b''
         try:
-            self.__openLog()
+            self.__openLog("fingerprint")
 
             # The fingerprint is always exectuted in a temporary directory
             tmpDir = tempfile.mkdtemp(dir=os.getcwd(), prefix=".bob-")
@@ -424,7 +482,8 @@ class Invoker:
 
             cmdArgs += self.__spec.language.setupFingerprint(self.__spec, env)
             proc = await self.__runCommand(cmdArgs, tmpDir, env=env,
-                universal_newlines=False, stdout=True, stderr=True)
+                universal_newlines=False, stdout=True, stderr=True,
+                specEnv=False)
             ret = proc.returncode
             stdout = proc.stdout
             stderr = proc.stderr
@@ -445,6 +504,36 @@ class Invoker:
 
         return (ret, stdout, stderr)
 
+    async def executeScmSwitch(self, scm, oldSpec):
+        # make permissions predictable
+        os.umask(0o022)
+
+        ret = 1
+        try:
+            self.__openLog("SCM inline switch")
+            try:
+                await scm.switch(self, oldSpec)
+            except CmdFailedError as e:
+                self.error(scm.getSource(), "failed")
+                self.error(e.what)
+                raise
+            except Exception:
+                self.error(scm.getSource(), "failed")
+                raise
+
+            # everything went well
+            ret = 0
+
+        except OSError as e:
+            self.error("Something went wrong:", str(e))
+            ret = 1
+        except InvocationError as e:
+            ret = e.returncode
+        finally:
+            self.__closeLog(ret)
+
+        return ret
+
     async def callCommand(self, args, cwd=None, **kwargs):
         cwd = os.path.join(self.__cwd, cwd) if cwd else self.__cwd
         ret = await self.__runCommand(args, cwd, **kwargs)
@@ -491,3 +580,6 @@ class Invoker:
         return self.__stdioBuffer.getvalue().decode(
             locale.getpreferredencoding(), 'surrogateescape')
 
+    def setMakeParameters(self, fds, jobs):
+        self.__makeFds = fds
+        self.__makeJobs = jobs
