@@ -96,22 +96,23 @@ class ImportScm(Scm):
         schema.Optional('prune') : bool,
     })
 
-    def __init__(self, spec, overrides=[]):
+    def __init__(self, spec, overrides=[], pruneDefault=None):
         super().__init__(spec, overrides)
         self.__url = spec["url"]
         self.__dir = spec.get("dir", ".")
-        self.__prune = spec.get("prune", False)
+        self.__prune = spec.get("prune", pruneDefault or False)
         self.__data = spec.get("__data")
 
-    def getProperties(self):
-        ret = super().getProperties()
+    def getProperties(self, isJenkins):
+        ret = super().getProperties(isJenkins)
         ret.update({
             'scm' : 'import',
             'url' : self.__url,
             'dir' : self.__dir,
             'prune' : self.__prune,
-            '__data' : packTree(self.__url),
         })
+        if isJenkins:
+            ret['__data'] = packTree(self.__url)
         return ret
 
     async def invoke(self, invoker):
@@ -134,6 +135,9 @@ class ImportScm(Scm):
     def isDeterministic(self):
         return False
 
+    def isLocal(self):
+        return True
+
     def hasLiveBuildId(self):
         return True
 
